import functools

from pyspark.sql import SparkSession

from sparksampling.config import SPARK_CONF
from sparksampling.mixin import WorkerManagerMixin, LogMixin


def record_job_id(method):
    """
    Decorate methods
    Set job_id to Engine, Currently supports spark
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        job_id = kwargs.get('job_id', getattr(self, 'job_id', None))
        if isinstance(self, SparkMixin) and job_id:
            self.log.info(f'Setting Spark job group: {job_id}')
            self.spark.sparkContext.setLogLevel('ERROR')
            self.spark.sparkContext.setJobGroup(job_id, f'Submitted by {self.__class__.__name__}',
                                                interruptOnCancel=True)
            # config for scheduler.mode: FAIR
            self.spark.sparkContext.setLocalProperty("spark.scheduler.pool", job_id)
        result = method(self, *args, **kwargs)
        return result

    return wrapper


class BaseEngine(WorkerManagerMixin):
    guarantee_worker = 10

    def submit(self, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def stop(cls, parent, job_id=None):
        raise NotImplementedError

    @classmethod
    def config(cls, kwargs):
        raise NotImplementedError

    @classmethod
    def is_matching(cls, request_type):
        return False


class SparkMixin(LogMixin):
    @property
    def spark(self) -> SparkSession:
        if not hasattr(self, 'parent'):
            return SparkSession.builder.config(conf=SPARK_CONF).getOrCreate()
        spark = getattr(self.parent, 'spark')
        if not spark:
            conf = getattr(self.parent, 'spark_config', SPARK_CONF)
            spark = SparkSession.builder.config(conf=conf).getOrCreate()
        return spark


class SparkBaseEngine(BaseEngine, SparkMixin):
    @classmethod
    def stop(cls, parent, job_id=None):
        # SparkStopEngine will cancel spark job
        return



import os
import socket

from kubernetes.config.incluster_config import SERVICE_TOKEN_FILENAME
from pyspark import SparkConf
from pyspark.sql import SparkSession

DEPLOY_ON_K8S = os.environ.get('DEPLOY_ON_K8S', 'false') in ['true', 'True']
INCLUSTER = os.path.exists('/var/run/secrets/kubernetes.io/serviceaccount/namespace')


def get_namespace():
    if not INCLUSTER:
        return 'spark-sampling'
    if os.getenv('SPARK_EXECUTOR_NS'):
        return os.getenv('SPARK_EXECUTOR_NS')
    with open('/var/run/secrets/kubernetes.io/serviceaccount/namespace') as f:
        default_namespace = f.read()
    return default_namespace


def get_k8s_config():
    from kubernetes import config
    from kubernetes.client import Configuration

    if INCLUSTER:
        config.load_incluster_config()
    else:
        config.load_kube_config('/root/.kube/config')
    k8s_config = Configuration.get_default_copy()
    token = k8s_config.api_key.get('authorization')
    host = k8s_config.host
    ca = k8s_config.ssl_ca_cert
    key_file = k8s_config.key_file
    cert_file = k8s_config.cert_file

    return host, token, ca, key_file, cert_file


def get_host_ip():
    # only use for minikube debug
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip


def set_if_not_none(config_dict, k, v):
    if v is not None:
        config_dict[k] = v


def get_platform_config():
    if DEPLOY_ON_K8S:
        url, token, ca, key_file, cert_file = get_k8s_config()
        master = f"k8s://{url}"
        image = os.getenv('SPARK_EXECUTOR_IMAGE', 'wh1isper/pyspark-executor:latest')

        label_names = os.getenv('SPARK_EXECUTOR_LABELS', '').split(',')
        label_names = (label for label in label_names if label)

        annotation_names = os.getenv('SPARK_EXECUTOR_ANNOTATIONS', '').split(',')
        annotation_names = (annotation for annotation in annotation_names if annotation)

        EXECUTOR_CONFIG = {
            **{f'spark.kubernetes.executor.label.{label_name}': 'true' for label_name in label_names},
            **{f'spark.kubernetes.executor.annotation.{annotation}': 'true' for annotation in annotation_names},
            "spark.executor.instances": os.getenv('SPARK_EXECUTOR_NUMS', '3'),
        }

        SCHEDULER_CONFIG = {
            "spark.master": master,
            "spark.kubernetes.authenticate.caCertFile": ca,
            "spark.kubernetes.authenticate.clientKeyFile": key_file,
            "spark.kubernetes.authenticate.clientCertFile": cert_file,
            "spark.kubernetes.container.image": image,
            "spark.kubernetes.container.image.pullPolicy": 'IfNotPresent',
            **EXECUTOR_CONFIG,
        }

        SCHEDULER_CONFIG = {k: v for k, v in SCHEDULER_CONFIG.items() if v is not None}
        if not INCLUSTER:
            set_if_not_none(SCHEDULER_CONFIG, 'spark.driver.host', get_host_ip())
        else:
            SCHEDULER_CONFIG['spark.kubernetes.authenticate.oauthTokenFile'] = SERVICE_TOKEN_FILENAME
            SCHEDULER_CONFIG['spark.driver.host'] = os.getenv('SPARK_DRIVER_HOST', 'spark-sampling-headless-service')
            SCHEDULER_CONFIG['spark.driver.bindAddress'] = os.getenv('SPARK_DRIVER_BINDADDRESS', '0.0.0.0')
            SCHEDULER_CONFIG['spark.kubernetes.driver.pod.name'] = os.getenv('SPARK_DRIVER_POD_NAME')

        set_if_not_none(SCHEDULER_CONFIG, 'spark.kubernetes.namespace', get_namespace())
    else:
        # as standalone
        SCHEDULER_CONFIG = {
            "spark.master": os.environ.get('SPARK_MASTER', 'local[*]'),
            "spark.driver.memory": "512m",
        }
    return SCHEDULER_CONFIG


def get_s3_config():
    AK = os.environ.get('S3_ACCESS_KEY', '')
    SK = os.environ.get('S3_SECRET_KEY', '')
    END_POINT = os.environ.get('S3_ENTRY_POINT', '')
    if not (AK and SK and END_POINT):
        return dict()
    return {
        'spark.hadoop.fs.s3a.access.key': AK,
        'spark.hadoop.fs.s3a.secret.key': SK,
        'spark.hadoop.fs.s3a.endpoint': END_POINT,
        "spark.hadoop.fs.s3a.path.style.access": os.environ.get('S3_PATH_STYLE_ACCESS', 'true'),

        # write s3 improvement
        'spark.hadoop.fs.s3a.bucket.all.committer.magic.enabled': 'true',
    }


def get_nfs_mount_config():
    volume_type = 'nfs'
    volume_name = 'driver-nfs-mount'
    mount_path = os.getenv('SPARK_EXECUTOR_VOLUNM_NFS_MOUNT_PATH')
    remote_server = os.getenv('SPARK_EXECUTOR_VOLUNM_NFS_REMOTE_SERVER')
    remote_path = os.getenv('SPARK_EXECUTOR_VOLUNM_NFS_REMOTE_PATH')
    if not (mount_path and remote_path and remote_server):
        return dict()

    return {
        f'spark.kubernetes.driver.volumes.{volume_type}.{volume_name}.mount.path': mount_path,
        f'spark.kubernetes.driver.volumes.{volume_type}.{volume_name}.mount.readOnly': 'false',
        f'spark.kubernetes.driver.volumes.{volume_type}.{volume_name}.options.server': remote_path,
        f'spark.kubernetes.driver.volumes.{volume_type}.{volume_name}.options.path': remote_path,
    }


def get_pvc_mount_config():
    pvc_name = os.getenv('SPARK_EXECUTOR_VOLUNM_PVC_NAME')
    pvc_path = os.getenv('SPARK_EXECUTOR_VOLUNM_PVC_PATH')
    if not (pvc_name and pvc_path):
        return dict()
    return {
        f'spark.kubernetes.executor.volumes.persistentVolumeClaim.{pvc_name}.mount.path': pvc_path,
        f'spark.kubernetes.executor.volumes.persistentVolumeClaim.{pvc_name}.mount.readOnly': 'false',
    }


def get_mount_config():
    return {
        **get_nfs_mount_config(),
        **get_pvc_mount_config(),
    }


def get_spark_conf():
    all_config = {
        "spark.submit.deployMode": 'client',
        **get_platform_config(),
        "spark.app.name": os.environ.get('SPARK_APP_NAME', 'Spark Sampling APP'),
        "spark.ui.port": os.environ.get('SPARK_UI_PORT', '12344'),
        **get_s3_config(),
        # scheduler as FAIR
        "spark.scheduler.mode": "FAIR",
    }

    config = []
    for k, v in all_config.items():
        config.append((k, v))

    spark_config = SparkConf().setAll(config)
    return spark_config


SPARK_CONF = get_spark_conf()


def get_spark():
    return SparkSession.builder.config(conf=SPARK_CONF).getOrCreate()


if __name__ == '__main__':
    spark = get_spark()



import os

from sparksampling.file_format.base_file_format import SparkBaseFileFormat
from sparksampling.file_format.output_adapter import OutputAdapterMixin


class CsvFileImpSpark(SparkBaseFileFormat, OutputAdapterMixin):
    cls_args = ['with_header', 'sep']

    def __init__(self, spark, *args, **kwargs):
        super(CsvFileImpSpark, self).__init__(spark, *args, **kwargs)
        self.with_header = kwargs.pop('with_header', True)
        default_sep = '\001' if not os.getenv('COMMA_SEP') else ','
        self.sep = kwargs.pop('sep', default_sep)

    def read(self, input_path):
        return self.spark.read.csv(input_path, sep=self.sep, header=self.with_header)

    def write(self, df, output_path):
        # When repartition(1), spark will write to a single file
        # Usually this is better for other applications, but there is a performance penalty
        if os.getenv('NO_REPARTITION'):
            df.write.csv(output_path, sep=self.sep, header=self.with_header, mode='overwrite')
        else:
            df.repartition(1).write.csv(output_path, sep=self.sep, header=self.with_header, mode='overwrite')

        return self._get_sampled_file(output_path, sep=self.sep, spark=self.spark)



import os

import findspark

findspark.init()

from concurrent import futures
from pyspark import SparkConf
from pyspark.sql import SparkSession
import logging
import grpc
from sparksampling.config import SPARK_CONF
from sparksampling.service import GRPCService
from traitlets.config import Application
from traitlets import (
    Integer,
    Unicode,
    Dict,
    default,
    Instance
)

_ONE_DAY_IN_SECONDS = 60 * 60 * 24

aliases = {
    'log-level': 'Application.log_level',
    'ip': 'SparkSamplingAPP.ip',
    'port': 'SparkSamplingAPP.port',
    'workers': 'SparkSamplingAPP.workers',
}

flags = {
    'debug': (
        {'Application': {'log_level': logging.DEBUG}},
        "set log level to logging.DEBUG (maximize logging output)",
    ),
}


class SparkSamplingAPP(Application):
    name = 'sparksampling'
    version = "0.1.0"
    description = """An application for starting a spark sampling server"""
    # the grpc server handle
    server = None

    # application config
    aliases = Dict(aliases)
    flags = Dict(flags)

    ip = Unicode(
        os.getenv('SERVICE_HOST', '0.0.0.0'), help="Host IP address for listening (default 0.0.0.0)."
    ).tag(config=True)

    port = Integer(
        int(os.getenv('SERVICE_PORT', 8530)), help="Port (default 8530)."
    ).tag(config=True)

    spark_config = Instance(SparkConf)
    spark = Instance(SparkSession)
    customer_engine_package = Unicode(
        os.getenv('SERVICE_CUSTOMER_ENGINE_PACKAGES', ''),
        help="Third part engine packages, like 'package_a.engine,package_b.engine'"
    ).tag(config=True)

    @default('spark_config')
    def _spark_config_default(self):
        return SPARK_CONF

    @default('spark')
    def _spark_default(self):
        return SparkSession.builder.config(conf=self.spark_config).getOrCreate()

    @default('log_level')
    def _log_level_default(self):
        return logging.DEBUG

    @default('log_datefmt')
    def _log_datefmt_default(self):
        """Exclude date from default date format"""
        return "%Y-%m-%d %H:%M:%S"

    @default('log_format')
    def _log_format_default(self):
        """override default log format to include time"""
        return "[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d] %(message)s"

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self.init_logger()
        self.init_spark()

    def init_logger(self):
        logger = logging.getLogger('sparksampling')
        logger.propagate = True
        logger.parent = self.log
        logger.setLevel(self.log.level)

    def init_spark(self):
        self.log.info(f"Started SparkSession, Spark version: {self.spark.version}")

    def start(self, argv=None):
        self.initialize(argv)
        self._add_server()
        self.server.add_insecure_port('%s:%d' % (self.ip, self.port))
        self.server.start()
        self.log.info("Spark Sampling Server Listening On %s:%s..." %
                      (self.ip, self.port))
        self.server.wait_for_termination()

    def _add_server(self):
        GRPCService.register_engine(self)
        engine_worker = GRPCService.get_worker_num()
        # need one reserve worker to reject request asap
        max_workers = engine_worker + 1
        self.log.info(f'Service allocate {engine_worker} workers for engine and 1 reserve for response')
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        GRPCService.add_to_server(self, self.server)

    @classmethod
    def launch(cls, argv=None):
        self = cls.instance()
        self.start(argv)


main = SparkSamplingAPP.launch

if __name__ == '__main__':
    main()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import sampling_service_pb2 as sampling__service__pb2


class SparkSamplingServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SamplingJob = channel.unary_unary(
                '/SparkSamplingService/SamplingJob',
                request_serializer=sampling__service__pb2.SamplingRequest.SerializeToString,
                response_deserializer=sampling__service__pb2.SamplingResponse.FromString,
                )
        self.RelationSamplingJob = channel.unary_unary(
                '/SparkSamplingService/RelationSamplingJob',
                request_serializer=sampling__service__pb2.RelationSamplingRequest.SerializeToString,
                response_deserializer=sampling__service__pb2.RelationSamplingResponse.FromString,
                )
        self.CancelJob = channel.unary_unary(
                '/SparkSamplingService/CancelJob',
                request_serializer=sampling__service__pb2.CancelRequest.SerializeToString,
                response_deserializer=sampling__service__pb2.CancelResponse.FromString,
                )


class SparkSamplingServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SamplingJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RelationSamplingJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SparkSamplingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SamplingJob': grpc.unary_unary_rpc_method_handler(
                    servicer.SamplingJob,
                    request_deserializer=sampling__service__pb2.SamplingRequest.FromString,
                    response_serializer=sampling__service__pb2.SamplingResponse.SerializeToString,
            ),
            'RelationSamplingJob': grpc.unary_unary_rpc_method_handler(
                    servicer.RelationSamplingJob,
                    request_deserializer=sampling__service__pb2.RelationSamplingRequest.FromString,
                    response_serializer=sampling__service__pb2.RelationSamplingResponse.SerializeToString,
            ),
            'CancelJob': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelJob,
                    request_deserializer=sampling__service__pb2.CancelRequest.FromString,
                    response_serializer=sampling__service__pb2.CancelResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'SparkSamplingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SparkSamplingService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SamplingJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SparkSamplingService/SamplingJob',
            sampling__service__pb2.SamplingRequest.SerializeToString,
            sampling__service__pb2.SamplingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RelationSamplingJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SparkSamplingService/RelationSamplingJob',
            sampling__service__pb2.RelationSamplingRequest.SerializeToString,
            sampling__service__pb2.RelationSamplingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CancelJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SparkSamplingService/CancelJob',
            sampling__service__pb2.CancelRequest.SerializeToString,
            sampling__service__pb2.CancelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from pyspark.sql import SparkSession, DataFrame

from sparksampling.mixin import LogMixin


class SparkBaseFileFormat(LogMixin):
    cls_args = []

    @classmethod
    def get_init_conf(cls, conf):
        if not cls.cls_args:
            cls.log.warning('No parameters to initialize?')
            return dict
        init_conf = dict()
        for arg in cls.cls_args:
            if conf.get(arg):
                init_conf[arg] = conf.get(arg)
        return init_conf

    def __init__(self, spark, *args, **kwarg):
        self.spark: SparkSession = spark

    def read(self, input_path):
        raise NotImplementedError

    def write(self, df: DataFrame, output_path):
        raise NotImplementedError
